/* jemalloc allocator interface for Hammer.
 * Copyright (C) 2012  Meredith L. Patterson, Dan "TQ" Hirsch
 * Copyright (C) 2020  Kia
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <setjmp.h>
#include <jemalloc/jemalloc.h>

#include "hammer.h"
#include "internal.h"



static void * h_arena_malloc_raw(HArena *arena, size_t size, bool need_zero);

void* h_alloc(HAllocator* mm__, size_t size) {
  void *p = mm__->alloc(mm__, size);
  if(!p)
    h_platform_errx(1, "memory allocation failed (%uB requested)\n", (unsigned int)size);
  return p;
}

HArena *h_new_arena(HAllocator* mm__, size_t block_size) {
  unsigned int arena_idx;

  struct HArena_ *ret = h_new(struct HArena_, 1);
  assert(ret != NULL);
  ret->mm__ = mm__;

  int rv;
  size_t sz = sizeof(arena_idx);
  rv = mallctl("arenas.create", &arena_idx, &sz, 0, 0);
//  printf("ar create RV is %d\n", rv);
  assert(rv == 0);
  printf("XXXXXXXXXXXX creating new arena, arena idx is %d\n", arena_idx);

  ret->arena_idx  = arena_idx;

  ret->is_arena_valid = 1;
  ret->except = NULL;
  return ret;
}

void h_arena_set_except(HArena *arena, jmp_buf *except)
{
  arena->except = except;
}

/*
static void *alloc_block(HArena *arena, size_t size)
{
  assert(arena->is_arena_valid == 1);

  void *block = mallocx(size, MALLOCX_TCACHE(arena->tcache_idx) | MALLOCX_ARENA(arena->arena_idx));
  if (!block) {
    if (arena->except)
      longjmp(*arena->except, 1);
    h_platform_errx(1, "memory allocation failed (%uB requested)\n", (unsigned int)size);
  }
  return block;
}
*/
void * h_arena_malloc_noinit(HArena *arena, size_t size) {
  return h_arena_malloc_raw(arena, size, false);
}

void * h_arena_malloc(HArena *arena, size_t size) {
  return h_arena_malloc_raw(arena, size, true);
}

static void * h_arena_malloc_raw(HArena *arena, size_t size,
                                 bool need_zero) {
  void *block;

  assert(arena->is_arena_valid == 1);

  if (need_zero) {
    block = mallocx(size, MALLOCX_TCACHE_NONE | MALLOCX_ARENA(arena->arena_idx) | MALLOCX_ZERO);
  }
  else {
    block = mallocx(size, MALLOCX_TCACHE_NONE | MALLOCX_ARENA(arena->arena_idx));
  }

  if (!block) {
    if (arena->except)
      longjmp(*arena->except, 1);
    h_platform_errx(1, "memory allocation failed (%uB requested)\n", (unsigned int)size);
  }

  assert(block != NULL);
//  printf("XXXX ARENA MALLOC for %p, %ld\n",block, size);

  return block;
}

void h_arena_free(HArena *arena, void* ptr) {
//  printf("XXXX ARENA actual FREE for %p\n",ptr);
  dallocx(ptr, MALLOCX_TCACHE_NONE);
}

void h_delete_arena(HArena *arena) {
  HAllocator *mm__ = arena->mm__;
  unsigned int arena_idx;
  int rv;
  size_t mib[3];
  size_t miblen;

  arena_idx  = arena->arena_idx;

  // now we destroy the arena.
  miblen = 3;
  mallctlnametomib("arena.0.destroy", mib, &miblen);
  mib[1] = arena_idx;
  //printf("mib[1] is %zu\n", mib[1]);

  rv = mallctlbymib(mib, miblen, NULL, NULL, NULL, 0);

  //printf("arena destroy RV is %d\n", rv);
  assert(rv == 0);

  h_free(arena);
  printf("XXXXXXXXXXXXXXXXXXXX arena destroyed succesfully, arena idx %d\n", arena_idx);
}

void h_allocator_stats(HArena *arena, HArenaStats *stats) {
  printf("THIS IS REALLY BAD XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
}
